#include "csidh.h"
#include "elligator.h"

static natural prod(std::vector<natural> x,long long from,long long to)
{
  if (to <= from) return natural(1);
  if (to == from + 1) return x[from];
  long long mid = (to + from) / 2;
  return prod(x,from,mid) * prod(x,mid,to);
}

static natural prod(std::vector<natural> x)
{
  return prod(x,0,x.size());
}

// inputs are reduced mod p
// each input is separately allowed to be 0, producing output 0
static void invert(std::vector<natural> &x,const mpz_class p,int W)
{
  if (x.size() == 0) return;
  if (x.size() == 1) {
    x[0] = x[0].invertmodprime(p,W);
    return;
  }

  std::vector<bit> iszero;
  iszero.resize(x.size());
  std::vector<natural> y;
  y.resize(x.size());
  for (long long i = 0;i < x.size();++i) {
    iszero[i] = x[i].isequal(0);
    x[i] = x[i] ^ iszero[i];
    if (i == 1) y[i-1] = (x[i-1]*x[i])%p;
    if (i >= 2) y[i-1] = (y[i-2]*x[i])%p;
  }
  y[x.size()-1] = y[x.size()-2].invertmodprime(p,W);

  for (long long i = x.size()-1;i >= 0;--i) {
    if (i >= 1) y[i-1] = (y[i]*x[i])%p;
    if (i >= 2) x[i] = (y[i]*y[i-2])%p;
    if (i == 1) x[i] = (y[i]*x[i-1])%p;
    if (i == 0) x[i] = y[i];
  }
  
  for (long long i = 0;i < x.size();++i)
    x[i] = x[i].mux(natural(0),iszero[i]);
}

std::tuple< curve, std::vector<natural> >
csidh_iteration(curve C,std::vector<mpz_class> factors,std::vector<natural> e,std::vector<bit> etwist,long long periteration,mpz_class counter)
{
  assert(periteration == 1 || periteration == 2);

  assert(factors.size() == e.size());
  if (factors.size() == 0) return std::make_tuple(C,e);

  mpz_class p = C.parent();
  int W = chain_bestW(p-2);
  mpz_class base = elligator_base(p,factors);

  for (long long i = 1;i < factors.size();++i)
    assert(factors[i-1] < factors[i]);
  mpz_class lmax = factors[factors.size()-1];
  mpz_class lmax2 = lmax;
  if (factors.size() >= 2) lmax2 = factors[factors.size()-2];

  std::vector<bit> thisl; thisl.resize(factors.size());
  std::vector<bit> thisl2; thisl2.resize(factors.size());
  natural l(1);
  natural l2(1);
  bit ltwist(0);
  std::vector<natural> cofactor;
  bit useresult(0);
  bit useresult2(0);

  if (periteration == 1) {
    bit wantl(1);
    for (long long i = factors.size() - 1;i >= 0;--i) {
      bit enonzero = ~e[i].isequal(0);
      thisl[i] = wantl & enonzero;
      l = l.mux(natural(factors[i]),thisl[i]);
      cofactor.push_back(natural(factors[i]).mux(natural(1),thisl[i]));
      ltwist ^= thisl[i] & etwist[i];
      wantl ^= thisl[i];
    }
    useresult = ~wantl;
  } else {
    bit wantl(1);
    bit wantl2(1);
    for (long long i = factors.size() - 1;i >= 0;--i) {
      bit enonzero = ~e[i].isequal(0);
      thisl[i] = wantl & enonzero;
      thisl2[i] = ~wantl & wantl2 & ~(ltwist ^ etwist[i]) & enonzero;
      l = l.mux(natural(factors[i]),thisl[i]);
      l2 = l2.mux(natural(factors[i]),thisl2[i]);
      cofactor.push_back(natural(factors[i]).mux(natural(1),thisl[i] | thisl2[i]));
      ltwist ^= thisl[i] & etwist[i];
      wantl ^= thisl[i];
      wantl2 ^= thisl2[i];
    }
    useresult = ~wantl;
    useresult2 = ~wantl2;
  }

  point_affine P = elligator(C,counter+1,ltwist,base);

  std::vector<natural> inpoints;

  if (periteration == 1) {
    point_proj Q = P * (prod(cofactor) << 2);
    point_affine Qaff = point_affine(C,Q.xcoordinate(W));

    auto result = Qaff.isogeny(l,lmax.get_si(),inpoints);

    bit orderl = std::get<0>(result);
    std::vector<natural> resultlist = std::get<1>(result);

    natural Anew = (resultlist[0] * resultlist[1].invertmodprime(p,W)) % p;
 
    useresult &= orderl;
 
    natural A = C.montgomeryA().mux(Anew,useresult);
    C = curve(p,A);
 
    for (long long i = factors.size() - 1;i >= 0;--i) {
      bit takeaway = useresult & thisl[i];
      e[i] = e[i].subtract(natural(takeaway),e[i].bits());
    }
  } else {
    // three cases being handled at once:
    // case 2: found l and l2
    //   compute R = ((p+1)/l l2)P
    //   compute l R, of order 1 or l2; compute isogeny
    //   apply isogeny to R, obtaining point of order 1 or l; compute isogeny
    // case 1: found l but not l2
    //   compute ((p+1)/l)P, hopefully of order l; compute isogeny
    // case 0: found neither l nor l2; do nothing

    point_proj R = P * (prod(cofactor) << 2);

    // First isogeny.
    point_proj Q = R * l;

    std::vector<natural> recip;
    natural QX = Q.numerator();
    natural QZ = Q.denominator();
    natural RX = R.numerator();
    natural RZ = R.denominator();
    recip.push_back(QZ);
    recip.push_back(RZ);
    invert(recip,p,W);
    natural Qx = (QX*recip[0]) % p;
    natural Rx = (RX*recip[1]) % p;

    inpoints.push_back(Rx);
    point_affine Qaff = point_affine(C,Qx);

    auto result = Qaff.isogeny(l2,lmax2.get_si(),inpoints);

    bit orderl2 = std::get<0>(result);
    std::vector<natural> resultlist = std::get<1>(result);

    recip.resize(0);
    recip.push_back(resultlist[1]);
    recip.push_back(resultlist[3]);
    invert(recip,p,W);
    natural Anew = (resultlist[0] * recip[0]) % p;
    natural Rnew = (resultlist[2] * recip[1]) % p;

    useresult2 &= orderl2;
    Rx = Rx.mux(Rnew,useresult2);

    natural A = C.montgomeryA().mux(Anew,useresult2);
    C = curve(p,A);

    // Second isogeny.
    point_affine Raff = point_affine(C,Rx);
    std::vector<natural> inpoints2;

    auto result2 = Raff.isogeny(l,lmax.get_si(),inpoints2);

    bit orderl = std::get<0>(result2);
    std::vector<natural> resultlist2 = std::get<1>(result2);

    Anew = (resultlist2[0] * resultlist2[1].invertmodprime(p,W)) % p;

    useresult &= orderl;

    A = C.montgomeryA().mux(Anew,useresult);
    C = curve(p,A);

    for (long long i = factors.size() - 1;i >= 0;--i) {
      bit takeaway = useresult & thisl[i];
      bit takeaway2 = useresult2 & thisl2[i];
      e[i] = e[i].subtract(natural(takeaway^takeaway2),e[i].bits());
    }
  }

  return std::make_tuple(C,e);
}

curve csidh(curve C,std::vector<mpz_class> factors,std::vector<natural> e,std::vector<bit> etwist,long long periteration,long long iterations)
{
  for (mpz_class counter = 1;iterations > 0;++counter,--iterations) {
    auto result = csidh_iteration(C,factors,e,etwist,periteration,counter);
    C = std::get<0>(result);
    e = std::get<1>(result);
  }
  return C;
}

