#ifndef point_h
#define point_h

#include <tuple>
#include "curve.h"

class point_affine;
class point_proj;

class point_proj {
  curve C;
  natural X;
  natural Z;

public:
  curve parent(void) const { return C; }
  natural numerator() const { return X; }
  natural denominator() const { return Z; }

  // nonzero W is required to be chain_bestW(p-2)
  natural xcoordinate(int W = 0) const {
    mpz_class p = C.parent();
    return (X * Z.invertmodprime(p,W)) % p;
  }

  point_proj(curve Cinit) : C(Cinit), X(0,0), Z(1,1) {
  }

  point_proj(curve Cinit,natural x) : C(Cinit), X(x), Z(1,1) {
  }

  point_proj(curve Cinit,natural X,natural Z) : C(Cinit), X(X), Z(Z) {
  }

  point_proj operator*(const natural n) const {
    mpz_class p = C.parent();
    natural A24 = C.montgomeryA() + natural(2);
    A24 = A24.minimod(p).halfmod(p).halfmod(p);

    natural X1 = X;
    natural Z1 = Z;
    natural X2(1,1);
    natural Z2(0,0);
    natural X3 = X;
    natural Z3 = Z;
    bit oldswap(0);
    
    for (long long i = n.bits() - 1;i >= 0;--i) {
      bit swap = n.bitatpos(i);
      bit control = swap ^ oldswap;
      X2.cswap(X3,control);
      Z2.cswap(Z3,control);
      oldswap = swap;

      // https://hyperelliptic.org/EFD/g1p/auto-montgom-xz.html#ladder-ladd-1987-m-3

      natural A = (X2+Z2).minimod(p);
      natural B = X2.subtractmod(Z2,p);
      natural C = (X3+Z3).minimod(p);
      natural D = X3.subtractmod(Z3,p);

      natural AA = A.square() % p;
      natural BB = B.square() % p;
      natural E = AA.subtractmod(BB,p);

      natural DA = (D*A) % p;
      natural CB = (C*B) % p;
      natural DAplusCB = (DA+CB).minimod(p);
      natural DAminusCB = DA.subtractmod(CB,p);

      natural X4 = AA*BB;
      natural Z4 = E*((BB + A24*E) % p);
      natural X5 = Z1*(DAplusCB.square() % p);
      natural Z5 = X1*(DAminusCB.square() % p);

      X4 %= p;
      Z4 %= p;
      X5 %= p;
      Z5 %= p;

      X2 = X4;
      Z2 = Z4;
      X3 = X5;
      Z3 = Z5;

    }

    X2.cswap(X3,oldswap);
    Z2.cswap(Z3,oldswap);

    return point_proj(C,X2,Z2);
  }
} ;

class point_affine {
  curve C;
  natural x;

public:
  curve parent(void) const { return C; }
  natural xcoordinate(void) const { return x; }

  point_affine(curve Cinit) : C(Cinit), x(0,0) {
  }

  point_affine(curve Cinit,natural xinit) : C(Cinit) {
    mpz_class p = C.parent();
    x = xinit % p;
  }

  point_proj operator*(const natural n) const {
    mpz_class p = C.parent();
    natural A24 = C.montgomeryA() + natural(2);
    A24 = A24.minimod(p).halfmod(p).halfmod(p);

    natural X1 = x;
    natural X2(1,1);
    natural Z2(0,0);
    natural X3 = x;
    natural Z3(1,1);
    bit oldswap(0);
    
    for (long long i = n.bits() - 1;i >= 0;--i) {
      bit swap = n.bitatpos(i);
      bit control = swap ^ oldswap;
      X2.cswap(X3,control);
      Z2.cswap(Z3,control);
      oldswap = swap;

      // https://hyperelliptic.org/EFD/g1p/auto-montgom-xz.html#ladder-ladd-1987-m-3

      natural A = (X2+Z2).minimod(p);
      natural B = X2.subtractmod(Z2,p);
      natural C = (X3+Z3).minimod(p);
      natural D = X3.subtractmod(Z3,p);

      natural AA = A.square() % p;
      natural BB = B.square() % p;
      natural E = AA.subtractmod(BB,p);

      natural DA = (D*A) % p;
      natural CB = (C*B) % p;
      natural DAplusCB = (DA+CB).minimod(p);
      natural DAminusCB = DA.subtractmod(CB,p);

      natural X4 = AA*BB;
      natural Z4 = E*((BB+A24*E) % p);
      natural X5 = DAplusCB.square();
      natural Z5 = X1*(DAminusCB.square() % p);

      X4 %= p;
      Z4 %= p;
      X5 %= p;
      Z5 %= p;

      X2 = X4;
      Z2 = Z4;
      X3 = X5;
      Z3 = Z5;
    }

    X2.cswap(X3,oldswap);
    Z2.cswap(Z3,oldswap);

    return point_proj(C,X2,Z2);
  }

  // caller guarantees: l is prime in 3, 5, 7, ..., lmax
  // caller guarantees: point has order 1 or l
  // if this point has order l: compute isogenous curve
  //   and return 1,A1,A0 where curve is A1/A0
  //   also, push some points through the isogeny
  //   returning X,Z for image of each point
  // if this point has order 1: return 0,...,...
  std::tuple< bit, std::vector<natural> >
  isogeny(natural l,long long lmax,std::vector<natural> points) const {
    bit orderl = ~x.isequal(0);

    assert(lmax >= 3);
    assert((lmax & 1) != 0);

    mpz_class p = C.parent();
    natural a = C.montgomeryA() + natural(2);
    a = a.minimod(p);
    natural d = C.montgomeryA().subtractmod(natural(2),p);
    natural A24 = a.halfmod(p).halfmod(p);

    std::vector<natural> Xi; // X-coord of i*point
    std::vector<natural> Zi; // Z-coord of i*point
    std::vector<natural> Yi; // X-Z, maybe not reduced
    std::vector<natural> Ti; // X+Z, maybe not reduced

    // will fill array positions 1,2,3,...,(lmax-1)/2
    Xi.resize((lmax+1)/2);
    Zi.resize((lmax+1)/2);
    Yi.resize((lmax+1)/2);
    Ti.resize((lmax+1)/2);

    Xi[1] = x;
    Zi[1] = natural(1);
    Yi[1] = x.subtractmod(natural(1),p);
    Ti[1] = x + natural(1);

    for (long long i = 2;i <= (lmax-1)/2;++i) {
      if ((i & 1) == 0) {
        // https://hyperelliptic.org/EFD/g1p/auto-montgom-xz.html#doubling-dbl-1987-m-3

        natural A = Ti[i/2];
        natural B = Yi[i/2];
        natural AA = A.square() % p;
        natural BB = B.square() % p;
        natural C = AA.subtractmod(BB,p);
        natural X3 = AA*BB;
        natural Z3 = C*((BB+A24*C) % p);
        X3 %= p;
        Z3 %= p;

        Xi[i] = X3;
        Zi[i] = Z3;
      } else {
        // https://hyperelliptic.org/EFD/g1p/auto-montgom-xz.html#diffadd-mdadd-1987-m

        natural A = Ti[i/2];
        natural B = Yi[i/2];
        natural C = Ti[i/2+1];
        natural D = Yi[i/2+1];
        natural DA = (D*A) % p;
        natural CB = (C*B) % p;
        natural DAplusCB = (DA+CB).minimod(p);
        natural DAminusCB = DA.subtractmod(CB,p);
        natural X5 = DAplusCB.square();
        natural Z5 = x*(DAminusCB.square() % p);
        X5 %= p;
        Z5 %= p;

        Xi[i] = X5;
        Zi[i] = Z5;
      }

      Yi[i] = Xi[i].subtractmod(Zi[i],p);
      Ti[i] = (Xi[i] + Zi[i]).minimod(p);
    }

    // as suggested in https://eprint.iacr.org/2018/782.pdf:
    // use 2016 moody--shumow formulas to find isogenous curve

    natural Yprod = Yi[1];
    natural Tprod = Ti[1];
    bit use = ~l.isequal(3);
    for (int i = 2;i <= (lmax-1)/2;++i) {
      natural Y2(1); Y2 = Y2.mux(Yi[i],use); Yprod = (Yprod * Y2) % p;
      natural T2(1); T2 = T2.mux(Ti[i],use); Tprod = (Tprod * T2) % p;
      use &= ~l.isequal(2*i+1);
    }

    natural anew = a.powpow8mod(l,Tprod,p);
    natural dnew = d.powpow8mod(l,Yprod,p);
    natural A1 = (anew+dnew).minimod(p).dbl().minimod(p);
    natural A0 = anew.subtractmod(dnew,p);

    std::vector<natural> result;
    result.push_back(A1);
    result.push_back(A0);

    // pushing points through:
    // https://eprint.iacr.org/2018/782.pdf mentions montgomery's speedup
    // but here points are affine, so use simpler formulas
    
    for (long long j = 0;j < points.size();++j) {
      natural Xout(1);
      natural Zout(1);
      bit use(1);
      for (int i = 1;i <= (lmax-1)/2;++i) {
        natural X2(1);
        X2 = X2.mux((points[j]*Xi[i]).subtractmod(Zi[i],p),use);
        Xout = (Xout * X2) % p;
        natural Z2(1);
        Z2 = Z2.mux((points[j]*Zi[i]).subtractmod(Xi[i],p),use);
        Zout = (Zout * Z2) % p;
        use &= ~l.isequal(2*i+1);
      }
      Xout = Xout.square() % p;
      Zout = Zout.square() % p;
      Xout = (Xout * points[j]) % p;
      result.push_back(Xout);
      result.push_back(Zout);
    }

    return std::make_tuple(orderl,result);
  }
} ;

#endif
